//---------------------------------------------------------------------------


#pragma hdrstop

#include "Process.h"
#include "Log.h"
#include <assert.h>

//---------------------------------------------------------------------------

#pragma package(smart_init)

namespace {

enum Context {
	CTX_IDLE = 0,
	CTX_STRING,
	CTX_STRING_ESCAPE,	// backslash inside string
	CTX_COMMENT_SLASH,	// after first slash starting comment
	CTX_COMMENT_S,	// comment: single line
	CTX_COMMENT_M,	// comment: multiline
	CTX_COMMENT_M_LEAVING	// inside multiline comment, after *, before slash
};

void EndComment(const AnsiString &in, int &pos, AnsiString &out, AnsiString &comment, TStrings *needles, bool removeAll,
				bool skipLongComments, unsigned int maxCommentLength
				)
{
	if (skipLongComments && (static_cast<unsigned int>(comment.Length()) > maxCommentLength))
	{
		out += comment;
		comment = "";		
		return;
	}

	if (removeAll == false)
	{
		for (int i=0; i<needles->Count; i++)
		{
			if (needles->Strings[i] != "")
			{
				if (comment.Pos(needles->Strings[i])) {
					// remove proceeding or following newline if empty line is left after removing comment
					int skipBackWs = 0;
					int backPos = pos - comment.Length() - 1;	// minus 1?
					bool prevNewline = false;
					while (backPos > 0)
					{
						char c = in[backPos--];
						if (c == '\n')
						{
							prevNewline = true;
							break;
						}
						else if (c == ' ' || c == '\t')
						{
							// skip whitespaces
							skipBackWs++;
						}
						else
						{
							break;
						}
					}
					if (backPos == 0)
					{
						// don't leave empty line at the beginning of the file
						prevNewline = true;
					}
					int frontPos = pos;
					bool nextNewline = false;
					int skipFrontWs = 0;
					while (frontPos <= in.Length())
					{
						char c = in[frontPos++];
						if (c == '\n')
						{
							nextNewline = true;
							break;
						}
						else if (c == ' ' || c == '\t' || c == '\r')
						{
							skipFrontWs++;
						}
						else
						{
							break;
						}
					}
					if (frontPos > in.Length())
					{
						// don't leave empty newline at the end of the file
						nextNewline = true;
					}

					if (prevNewline && nextNewline)
					{
						if (backPos == 0)
						{
							out = out.SubString(1, out.Length() - skipBackWs);
							pos += skipFrontWs + 1;
						}
						else if (frontPos > in.Length())
						{
							out = out.SubString(1, out.Length() - skipBackWs - 1);
							if (out[out.Length()] == '\r')
							{
								out = out.SubString(1, out.Length() - 1);
							}
							pos += skipFrontWs;
						}
						else
						{
							out = out.SubString(1, out.Length() - skipBackWs);
							pos += skipFrontWs + 1;
						}
					}

					comment = "";
					return;
				}
			}
		}
		out += comment;
	}
	comment = "";
}


}	// namespace


int ProcessCpp(const AnsiString &in, TStrings *needles, bool removeAll,
				bool skipLongComments, unsigned int maxCommentLength,
				AnsiString &out)
{
	if (needles->Count == 0 && removeAll == false)
	{
		out = in;
		return -1;
	}

	int pos = 1;
	int length = in.Length();
	enum Context ctx = CTX_IDLE;
	AnsiString commentBuf;
	while (pos <= length)
	{
		char c = in[pos++];
		
		switch (ctx)
		{

		case CTX_IDLE: {
			switch (c) {
			case '"':
				ctx = CTX_STRING;
				out += c;
				break;
			case '/':
				ctx = CTX_COMMENT_SLASH;
				commentBuf += c;
				break;
			default:
				out += c;
				break;
			}
		}
		break;

		case CTX_STRING: {
			switch (c) {
			case '\\':
				ctx = CTX_STRING_ESCAPE;
				break;
			case '"':
				ctx = CTX_IDLE;
				break;
			default:
				break;
			}
			out += c;			
		}
		break;

		case CTX_STRING_ESCAPE: {
			out += c;
			ctx = CTX_STRING;
		}
		break;

		case CTX_COMMENT_SLASH: {
			switch (c) {
			case '*':
				ctx = CTX_COMMENT_M;
				commentBuf += c;
				break;
			case '/':
				ctx = CTX_COMMENT_S;
				commentBuf += c;
				break;
			default:
				out += commentBuf;
				commentBuf = "";
				out += c;
				ctx = CTX_IDLE;
				break;
			}
		}
		break;

		case CTX_COMMENT_S: {
			switch (c) {
			case '\r':
			/*
				commentBuf += c;
				if (pos <= length)
				{
					c = in[pos++];
					commentBuf += c;
				}
			*/
				pos--;
				EndComment(in, pos, out, commentBuf, needles, removeAll, skipLongComments, maxCommentLength);
				ctx = CTX_IDLE;
				break;
			case '\n':
			//	commentBuf += c;
				pos--;
				EndComment(in, pos, out, commentBuf, needles, removeAll, skipLongComments, maxCommentLength);
				ctx = CTX_IDLE;
				break;
			default:
				commentBuf += c;
				break;
			}
		}
		break;

		case CTX_COMMENT_M: {
			commentBuf += c;
			switch (c) {
			case '*':
				ctx = CTX_COMMENT_M_LEAVING;
				break;
			default:
				break;
			}
		}
		break;

		case CTX_COMMENT_M_LEAVING: {
			commentBuf += c;
			switch (c) {
			case '*':
				ctx = CTX_COMMENT_M_LEAVING;
				break;
			case '/':
				EndComment(in, pos, out, commentBuf, needles, removeAll, skipLongComments, maxCommentLength);
				ctx = CTX_IDLE;
				break;
			default:
				ctx = CTX_COMMENT_M;
				break;
			}
		}
		break;

		default:
			assert(!"Unhandled context!");
			break;

        }
	}

	if (ctx == CTX_COMMENT_S)
	{
		EndComment(in, pos, out, commentBuf, needles, removeAll, skipLongComments, maxCommentLength);
	}
	else if (ctx != CTX_IDLE)
	{
    	LOG("Unexpected: context = %d at the end of the file\n", ctx);
    }

	return 0;
}

